// $Date:: 2010-02-21 22:22:21 +0900#$
// $Rev: 65 $

#define WIN32_LEAN_AND_MEAN

#include "PNLM.h"
#include <math.h>
#include <smmintrin.h>
#include <windows.h>

//#define PRINTTIME
#ifdef PRINTTIME
#include <stdio.h>
#include <mmsystem.h>
#pragma comment ( lib , "winmm.lib" )
static __int32 tcount = 0 ;
#endif

// m 4 oCgɃpbLO
#pragma pack(push,4)
typedef union
{
    __int32 int32 ;
    __int16 int16 [ 2 ] ;
} _weight ;
#pragma pack(pop)

static DWORD WINAPI ThreadFunc ( LPVOID lpParameter ) ;
static void SetCacheFrame_YV12_SSE2 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM& p ) ;
static void SetCacheFrame_YV12_SSE4 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM& p ) ;
static void SetCacheFrame_YUY2_SSE2 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM& p ) ;
static void SetCacheFrame_YUY2_SSE4 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM& p ) ;
static void SetFrame_YV12_SSE2 ( THREAD_PARAM& p , int i , int j , __m128i& s ) ;
static void SetFrame_YV12_SSE4 ( THREAD_PARAM& p , int i , int j , __m128i& s ) ;
static void SetFrame_YUY2_SSE2 ( THREAD_PARAM& p , int i , int j , __m128i& s ) ;
static void SetFrame_YUY2_SSE4 ( THREAD_PARAM& p , int i , int j , __m128i& s ) ;
static DWORD WINAPI ThreadFunc_SSE2 ( LPVOID lpParameter ) ;
static DWORD WINAPI ThreadFunc_SSE2_3x3 ( LPVOID lpParameter ) ;
static DWORD WINAPI ThreadFunc_SSE2_5x5 ( LPVOID lpParameter ) ;
static DWORD WINAPI ThreadFunc_SSE4 ( LPVOID lpParameter ) ;
static DWORD WINAPI ThreadFunc_SSE4_3x3 ( LPVOID lpParameter ) ;
static DWORD WINAPI ThreadFunc_SSE4_5x5 ( LPVOID lpParameter ) ;

/**************************************************************************************************/
/* PNLM                                                                                      */
/**************************************************************************************************/
//==============================================================================
// constructor
//==============================================================================
PNLM::PNLM(PClip _child, int _strength, int _wspan , int _tspan, int _threads, int _opt, IScriptEnvironment* env) :
    GenericVideoFilter(_child),
    Strength ( _strength) ,
    wSpan ( _wspan ) ,
    tSpan ( _tspan ) ,
    Threads ( _threads )
{
    this->Height    = this->vi.height ;
    this->Width     = this->vi.width ;

	if( _strength < 0 || 110 < _strength )
		env->ThrowError("PNLM:'STRENGTH' must be 0 to 110.");

	if( _wspan < 0 || 15 < _wspan )
		env->ThrowError("PNLM:'WSPAN' must be 0 to 15.");

	if( _tspan < 0 || 15 < _tspan )
		env->ThrowError("PNLM:'TSPAN' must be 0 to 15.");

	if( _threads < 1 || 15 < _threads )
		env->ThrowError("PNLM:'THREADS' must be 1 to 15.");

	if( _opt < 0 || 2 < _opt )
		env->ThrowError("PNLM:'OPT' must be 0 to 2.");

    if( _opt == 0 )
    {
	    unsigned int output_ecx;
	    unsigned int output_edx;
	    __asm{
		    push eax
		    push ebx
		    push ecx
		    push edx

		    mov eax, 1

		    CPUID

		    mov output_ecx, edx
		    mov output_edx, edx

		    pop edx
		    pop ecx
		    pop ebx
		    pop eax
	    }

	    if ( ! ( output_edx & 0x04000000 ) )
		    env->ThrowError("PNLM:this machine do not support SSE2.");

        _opt = ( output_ecx & 0x00080000 ) ? 2 : 1 ;
    }


    // Zx->mCYlϊs̐B
    const double H = - pow ( 10, (double) _strength / -22.0) ;
    this->pW = ( (__int32*) _aligned_malloc ( 512 * sizeof (__int32) , 16 ) ) + 256 ;
    for ( int i = -255 ; i < 256 ; i ++ )
    {
        _weight w ;
        w.int16 [ 0 ] = (__int16) ( exp ( H * i * i ) * 32 + 0.5 );
        w.int16 [ 1 ] = (__int16) ( exp ( H * i * i ) * 8 * i + 0.5 );
        this->pW [ i ] = w.int32 ;
    }

    // 炵
    this->CacheLOffset = ( ( ( this->wSpan - 1 ) >> 2 ) + 1 ) << 2 ;

    // PC̃oCg
    this->CacheLSpan = this->CacheLOffset + ( ( ( ( this->Width + this->wSpan - 1 ) >> 2 ) + 1 ) << 2 ) ;

    // LbVobt@
    this->FrameCache = ( (__int32**) malloc ( ( 2 * this->tSpan + 1 ) * sizeof (__int32*) ) ) + this->tSpan ;
    for ( int i = -this->tSpan ; i <= this->tSpan  ; i ++ )
    {
        this->FrameCache [ i ] = (__int32*) _aligned_malloc ( this->CacheLSpan * ( this->Height + 2 * this->wSpan ) * sizeof (__int32) , 16 ) ;
    }

    this->PreFrameNum = -2 ;

    //this->ThreadFunc = ::ThreadFunc_SSE4_5x5 ;
    this->ThreadFunc = ( _opt == 1 ) ? ( ( this->wSpan == 1 ) ? ::ThreadFunc_SSE2_3x3 : ( ( this->wSpan == 2 ) ? ::ThreadFunc_SSE2_5x5 : ::ThreadFunc_SSE2 ) )
                                     : ( ( this->wSpan == 1 ) ? ::ThreadFunc_SSE4_3x3 : ( ( this->wSpan == 2 ) ? ::ThreadFunc_SSE4_5x5 : ::ThreadFunc_SSE4 ) ) ;
    this->SetCacheFrame = ( _opt == 1 ) ? ( this->vi.IsYV12 () ? ::SetCacheFrame_YV12_SSE2 : ::SetCacheFrame_YUY2_SSE2 )
                                        : ( this->vi.IsYV12 () ? ::SetCacheFrame_YV12_SSE4 : ::SetCacheFrame_YUY2_SSE4 ) ;
    this->SetFrame = ( _opt == 1 ) ? ( this->vi.IsYV12 () ? ::SetFrame_YV12_SSE2 : ::SetFrame_YUY2_SSE2 )
                                   : ( this->vi.IsYV12 () ? ::SetFrame_YV12_SSE4 : ::SetFrame_YUY2_SSE4 ) ;
}


//==============================================================================
// destructor
//==============================================================================
PNLM::~PNLM()
{
    for ( int i = -this->tSpan ; i <= this->tSpan  ; i ++ )
    {
        if ( this->FrameCache [ i ] ) _aligned_free ( this->FrameCache [ i ] ) ;
    }
    free ( this->FrameCache - this->tSpan ) ;
    _aligned_free ( this->pW - 256 ) ;

#ifdef PRINTTIME
    char strtime [ 256 ] ;
    _snprintf ( strtime , sizeof (strtime) , "%f\r\n" , ( (float) tcount ) / 1000 ) ;
    OutputDebugString ( strtime ) ;
#endif // PRINTTIME
}


//==============================================================================
// GetFrame
//==============================================================================
PVideoFrame PNLM::GetFrame (int n, IScriptEnvironment* env)
{
    PVideoFrame src = this->child->GetFrame ( n , env ) ;
    env->MakeWritable ( & src ) ;

    BYTE* pSrcY = src->GetWritePtr () ;

#ifdef PRINTTIME
    tcount -= timeGetTime () ;
#endif // PRINTTIME

    if ( this->PreFrameNum == -2 )
    {
        // ߂ GetFrame Ă΂ꂽƂ
        this->SrcYPitch = src->GetPitch() ;
    }

    THREAD_PARAM p (
        this->pW ,
        this->wSpan ,
        this->tSpan ,
        this->Threads ,
        this->SrcYPitch ,
        this->Height ,
        this->Width ,
        this->CacheLSpan ,
        this->CacheLOffset ,
        this->FrameCache ,
        pSrcY ,
        1 ,
        this->SetFrame
        ) ;
    
    if ( this->PreFrameNum + 1 != n )
    {
        for ( int i = -this->tSpan ; i < this->tSpan  ; i ++ )
        {
            if ( n + i < 0 )
            {
                this->SetCacheFrame ( i , 0 , env , this->child , p ) ;
            }
            else if ( n + i >= this->vi.num_frames )
            {
                this->SetCacheFrame ( i , this->vi.num_frames - 1 , env , this->child , p ) ;
            }
            else
            {
                this->SetCacheFrame ( i , n + i , env , this->child , p ) ;
            }
        }
    }
    else
    {
        __int32* pTmp = this->FrameCache [ -this->tSpan ] ;
        memcpy ( this->FrameCache - this->tSpan , this->FrameCache - this->tSpan + 1 ,
            2 * this->tSpan * sizeof (BYTE*) ) ;
        this->FrameCache [ this->tSpan ] = pTmp ;
    }
    this->PreFrameNum = n ;

    if ( n + this->tSpan >= this->vi.num_frames )
    {
        this->SetCacheFrame ( this->tSpan , this->vi.num_frames - 1 , env , this->child , p ) ;
    }
    else
    {
        this->SetCacheFrame ( this->tSpan , n + this->tSpan , env , this->child , p ) ;
    }

#if 0
    this->ThreadFunc ( & p ) ;
#else
    HANDLE * hThr = (HANDLE*) alloca ( this->Threads * sizeof (HANDLE) ) ;
    THREAD_PARAM * sParam = (THREAD_PARAM*) alloca ( this->Threads * sizeof (THREAD_PARAM) ) ;
    for ( int i =0 ; i < this->Threads ; i ++ )
    {
        memcpy ( sParam + i , & p , sizeof (THREAD_PARAM) ) ;
        sParam [ i ].thr_num = i ;
    }
    for ( int i = 0 ; i < this->Threads ; i ++ )
        hThr [ i ] = ::CreateThread ( NULL , 0 , this->ThreadFunc , sParam + i , 0 , NULL ) ;
    for ( int i = 0 ; i < this->Threads ; i ++ )
        ::WaitForSingleObject ( hThr [ i ] , INFINITE ) ;
#endif

#ifdef PRINTTIME
    tcount += timeGetTime () ;
#endif // PRINTTIME
    return src ;
}


//==============================================================================
// SetCacheFrame
//==============================================================================
static void SetCacheFrame_YV12_SSE2 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM& p )
{
    __int32 * ppp = p.FrameCache [ cache_idx ] + p.wSpan * p.CacheLSpan ;
    const BYTE * src = child->GetFrame ( frame_num , env )->GetReadPtr () ;
    const BYTE * c = src ;
    for ( int j = 0 ; j < p.Height ; j ++ )
    {
        const BYTE * sss = src + j * p.SrcYPitch ;
        const int www = ( ( p.Width - 1 ) & ~ 0x03 ) + 4 ;

        // [
        for ( int i = 0 ; i < p.CacheLOffset >> 2 ; i ++ , ppp += 4 )
            _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi32 ( (* sss) << 2 ) ) ;

        // gi16 oCgŏł镪j
        for ( int i = 0 ; i < ( p.Width & ~ 0x0F ) ; i += 16 , ppp += 16 , sss += 16 )
        {
            const __m128i qqq = _mm_loadu_si128 ( (__m128i*) sss ) ;
            const __m128i rrr0 = _mm_slli_epi16 ( _mm_unpacklo_epi8 ( qqq , _mm_setzero_si128 () ) , 2 ) ;
            const __m128i rrr1 = _mm_slli_epi16 ( _mm_unpackhi_epi8 ( qqq , _mm_setzero_si128 () ) , 2 ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 0  ) , _mm_unpacklo_epi16 ( rrr0 , _mm_setzero_si128 () ) ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 4  ) , _mm_unpackhi_epi16 ( rrr0 , _mm_setzero_si128 () ) ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 8  ) , _mm_unpacklo_epi16 ( rrr1 , _mm_setzero_si128 () ) ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 12 ) , _mm_unpackhi_epi16 ( rrr1 , _mm_setzero_si128 () ) ) ;
        }
        // ci16 oCgɖȂj
        for ( int i = 0 ; i < ( p.Width & 0x0F ) ; i ++ , ppp ++ , sss ++ )
            * ppp = (* sss) << 2 ;

        // E[
        const __m128i last = _mm_set1_epi32 ( sss [ -1 ] << 2 ) ;
        CopyMemory ( ppp , & last , ( www - p.Width ) * sizeof (__int32) ) ;
        ppp += www - p.Width ;
        for ( int i = 0 ; i < ( p.CacheLSpan - p.CacheLOffset - www ) >> 2 ; i ++ , ppp +=4 )
            _mm_store_si128 ( (__m128i*) ppp , last ) ;
    }

    const __int32 * qqq ;

    // [
    qqq = p.FrameCache [ cache_idx ] + ( p.Height + p.wSpan - 1 ) * p.CacheLSpan ;
    ppp = p.FrameCache [ cache_idx ] + ( p.Height + p.wSpan ) * p.CacheLSpan ;
    for ( int i = 0 ; i < p.wSpan ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int32) ) ;

    // [
    qqq = p.FrameCache [ cache_idx ] + p.wSpan * p.CacheLSpan ;
    ppp = p.FrameCache [ cache_idx ] ;
    for ( int i = 0 ; i < p.wSpan ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int32) ) ;
}

static void SetCacheFrame_YV12_SSE4 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM& p )
{
    __int32 * ppp = p.FrameCache [ cache_idx ] + p.wSpan * p.CacheLSpan ;
    __int32 * qqq ;
    const BYTE * src = child->GetFrame ( frame_num , env )->GetReadPtr () ;
    const BYTE * c = src ;
    for ( int j = 0 ; j < p.Height ; j ++ )
    {
        const BYTE * sss = src + j * p.SrcYPitch ;
        const int www = ( ( p.Width - 1 ) & ~ 0x03 ) + 4 ;

        // [
        for ( int i = 0 ; i < p.CacheLOffset >> 2 ; i ++ , ppp += 4 )
            _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi32 ( (* sss) << 2 ) ) ;

        // gi16 oCgŏł镪j
        for ( int i = 0 ; i < ( p.Width & ~ 0x0F ) ; i += 16 , ppp += 16 , sss += 16 )
        {
            const __m128i wk = _mm_lddqu_si128 ( (__m128i*) sss ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 0  ) , _mm_slli_epi16 ( _mm_cvtepu8_epi32 ( wk                         ) , 2 ) ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 4  ) , _mm_slli_epi16 ( _mm_cvtepu8_epi32 ( _mm_srli_si128 ( wk , 4  ) ) , 2 ) ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 8  ) , _mm_slli_epi16 ( _mm_cvtepu8_epi32 ( _mm_srli_si128 ( wk , 8  ) ) , 2 ) ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 12 ) , _mm_slli_epi16 ( _mm_cvtepu8_epi32 ( _mm_srli_si128 ( wk , 12 ) ) , 2 ) ) ;
        }
        // ci16 oCgɖȂj
        for ( int i = 0 ; i < ( p.Width & 0x0F ) ; i ++ , ppp ++ , sss ++ )
            * ppp = (* sss) << 2 ;

        // E[
        const __m128i last = _mm_set1_epi32 ( sss [ -1 ] << 2 ) ;
        CopyMemory ( ppp , & last , ( www - p.Width ) * sizeof (__int32) ) ;
        ppp += www - p.Width ;
        for ( int i = 0 ; i < ( p.CacheLSpan - p.CacheLOffset - www ) >> 2 ; i ++ , ppp +=4 )
            _mm_store_si128 ( (__m128i*) ppp , last ) ;
    }

    // [
    qqq = p.FrameCache [ cache_idx ] + ( p.Height + p.wSpan - 1 ) * p.CacheLSpan ;
    ppp = p.FrameCache [ cache_idx ] + ( p.Height + p.wSpan ) * p.CacheLSpan ;
    for ( int i = 0 ; i < p.wSpan ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int32) ) ;

    // [
    qqq = p.FrameCache [ cache_idx ] + p.wSpan * p.CacheLSpan ;
    ppp = p.FrameCache [ cache_idx ] ;
    for ( int i = 0 ; i < p.wSpan ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int32) ) ;
}


static void SetCacheFrame_YUY2_SSE2 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM& p )
{
    __int32 * ppp = p.FrameCache [ cache_idx ] + p.wSpan * p.CacheLSpan ;
    const BYTE * src = child->GetFrame ( frame_num , env )->GetReadPtr () ;
    const BYTE * c = src ;
    const __m128i mask = _mm_set1_epi16 ( 0x0FF ) ;
    for ( int j = 0 ; j < p.Height ; j ++ )
    {
        const BYTE * sss = src + j * p.SrcYPitch ;
        const int www = ( ( p.Width - 1 ) & ~ 0x03 ) + 4 ;

        // [
        for ( int i = 0 ; i < p.CacheLOffset >> 2 ; i ++ , ppp += 4 )
            _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi32 ( (* sss) << 2 ) ) ;

        // gi16 oCgŏł镪j
        for ( int i = 0 ; i < ( p.Width & ~ 0x07 ) ; i += 8 , ppp += 8 , sss += 16 )
        {
            const __m128i qqq = _mm_slli_epi16 ( _mm_and_si128 ( _mm_loadu_si128 ( (__m128i*) sss ) , mask ) , 2 ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 0  ) , _mm_unpacklo_epi16 ( qqq , _mm_setzero_si128 () ) ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 4  ) , _mm_unpackhi_epi16 ( qqq , _mm_setzero_si128 () ) ) ;
        }
        // ci16 oCgɖȂj
        for ( int i = 0 ; i < ( p.Width & 0x07 ) ; i ++ , ppp ++ , sss += 2 )
            * ppp = (* sss) << 2 ;

        // E[
        const __m128i last = _mm_set1_epi32 ( sss [ -2 ] << 2 ) ;
        CopyMemory ( ppp , & last , ( www - p.Width ) * sizeof (__int32) ) ;
        ppp += www - p.Width ;
        for ( int i = 0 ; i < ( p.CacheLSpan - p.CacheLOffset - www ) >> 2 ; i ++ , ppp +=4 )
            _mm_store_si128 ( (__m128i*) ppp , last ) ;
    }

    const __int32 * qqq ;

    // [
    qqq = p.FrameCache [ cache_idx ] + ( p.Height + p.wSpan - 1 ) * p.CacheLSpan ;
    ppp = p.FrameCache [ cache_idx ] + ( p.Height + p.wSpan ) * p.CacheLSpan ;
    for ( int i = 0 ; i < p.wSpan ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int32) ) ;

    // [
    qqq = p.FrameCache [ cache_idx ] + p.wSpan * p.CacheLSpan ;
    ppp = p.FrameCache [ cache_idx ] ;
    for ( int i = 0 ; i < p.wSpan ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int32) ) ;
}

static void SetCacheFrame_YUY2_SSE4 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM& p )
{
    __int32 * ppp = p.FrameCache [ cache_idx ] + p.wSpan * p.CacheLSpan ;
    __int32 * qqq ;
    const BYTE * src = child->GetFrame ( frame_num , env )->GetReadPtr () ;
    const BYTE * c = src ;
    const __m128i mask = _mm_set1_epi16 ( 0x0FF ) ;
    for ( int j = 0 ; j < p.Height ; j ++ )
    {
        const BYTE * sss = src + j * p.SrcYPitch ;
        const int www = ( ( p.Width - 1 ) & ~ 0x03 ) + 4 ;

        // [
        for ( int i = 0 ; i < p.CacheLOffset >> 2 ; i ++ , ppp += 4 )
            _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi32 ( (* sss) << 2 ) ) ;

        // gi16 oCgŏł镪j
        for ( int i = 0 ; i < ( p.Width & ~ 0x07 ) ; i += 8 , ppp += 8 , sss += 16 )
        {
            const __m128i wk = _mm_slli_epi16 ( _mm_and_si128 ( _mm_lddqu_si128 ( (__m128i*) sss ) , mask ) , 2 ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 0  ) , _mm_cvtepu16_epi32 ( wk                         ) ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 4  ) , _mm_cvtepu16_epi32 ( _mm_srli_si128 ( wk , 8  ) ) ) ;
        }
        // ci16 oCgɖȂj
        for ( int i = 0 ; i < ( p.Width & 0x07 ) ; i ++ , ppp ++ , sss ++ )
            * ppp = (* sss) << 2 ;

        // E[
        const __m128i last = _mm_set1_epi32 ( sss [ -1 ] << 2 ) ;
        CopyMemory ( ppp , & last , ( www - p.Width ) * sizeof (__int32) ) ;
        ppp += www - p.Width ;
        for ( int i = 0 ; i < ( p.CacheLSpan - p.CacheLOffset - www ) >> 2 ; i ++ , ppp +=4 )
            _mm_store_si128 ( (__m128i*) ppp , last ) ;
    }

    // [
    qqq = p.FrameCache [ cache_idx ] + ( p.Height + p.wSpan - 1 ) * p.CacheLSpan ;
    ppp = p.FrameCache [ cache_idx ] + ( p.Height + p.wSpan ) * p.CacheLSpan ;
    for ( int i = 0 ; i < p.wSpan ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int32) ) ;

    // [
    qqq = p.FrameCache [ cache_idx ] + p.wSpan * p.CacheLSpan ;
    ppp = p.FrameCache [ cache_idx ] ;
    for ( int i = 0 ; i < p.wSpan ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int32) ) ;
}


//==============================================================================
// SetFrame
//==============================================================================
static void SetFrame_YV12_SSE2 ( THREAD_PARAM& p , int i , int j , __m128i& s )
{
    __m128i * pSet = (__m128i *) ( p.pSrcY + i + j * p.SrcYPitch ) ;
    __m128i s1 = _mm_unpacklo_epi8 ( _mm_loadu_si128 ( pSet ) , _mm_setzero_si128 () ) ;
    __m128i s2 = _mm_srai_epi16 ( s , 2 ) ;
    s1.m128i_i16 [ 0 ] += s.m128i_i16 [ 1 ] / s2.m128i_i16 [ 0 ] ;
    s1.m128i_i16 [ 1 ] += s.m128i_i16 [ 3 ] / s2.m128i_i16 [ 2 ] ;
    s1.m128i_i16 [ 2 ] += s.m128i_i16 [ 5 ] / s2.m128i_i16 [ 4 ] ;
    s1.m128i_i16 [ 3 ] += s.m128i_i16 [ 7 ] / s2.m128i_i16 [ 6 ] ;
    CopyMemory ( pSet , _mm_packus_epi16 ( s1 , _mm_setzero_si128 () ).m128i_u8 , 4 ) ;
}

static void SetFrame_YV12_SSE4 ( THREAD_PARAM& p , int i , int j , __m128i& s )
{
    __m128i * pSet = (__m128i *) ( p.pSrcY + i + j * p.SrcYPitch ) ;
    __m128i s1 = _mm_cvtepu8_epi16 ( _mm_lddqu_si128 ( pSet ) ) ;
    __m128i s2 = _mm_srai_epi16 ( s , 2 ) ;
    s1.m128i_i16 [ 0 ] += ( (__int16) _mm_extract_epi16 ( s , 1 ) ) / _mm_extract_epi16 ( s2 , 0 ) ;
    s1.m128i_i16 [ 1 ] += ( (__int16) _mm_extract_epi16 ( s , 3 ) ) / _mm_extract_epi16 ( s2 , 2 ) ;
    s1.m128i_i16 [ 2 ] += ( (__int16) _mm_extract_epi16 ( s , 5 ) ) / _mm_extract_epi16 ( s2 , 4 ) ;
    s1.m128i_i16 [ 3 ] += ( (__int16) _mm_extract_epi16 ( s , 7 ) ) / _mm_extract_epi16 ( s2 , 6 ) ;
    CopyMemory ( pSet , _mm_packus_epi16 ( s1 , _mm_setzero_si128 () ).m128i_u8 , 4 ) ;
}

static void SetFrame_YUY2_SSE2 ( THREAD_PARAM& p , int i , int j , __m128i& s )
{
    __m128i * pSet = (__m128i *) ( p.pSrcY + 2 * i + j * p.SrcYPitch ) ;
    __m128i s0 = _mm_loadu_si128 ( pSet ) ;
    __m128i s1 = _mm_srai_epi16 ( s , 2 ) ;
    _mm_storeu_si128 ( pSet , _mm_add_epi8 ( _mm_and_si128 ( _mm_setr_epi16 (
                    s.m128i_i16 [ 1 ] / s1.m128i_i16 [ 0 ] ,
                    s.m128i_i16 [ 3 ] / s1.m128i_i16 [ 2 ] ,
                    s.m128i_i16 [ 5 ] / s1.m128i_i16 [ 4 ] ,
                    s.m128i_i16 [ 7 ] / s1.m128i_i16 [ 6 ] ,
                    0 , 0 , 0 , 0 ) , _mm_set1_epi16 ( 0x0FF )) , s0 ) ) ;
}

static void SetFrame_YUY2_SSE4 ( THREAD_PARAM& p , int i , int j , __m128i& s )
{
    __m128i * pSet = (__m128i *) ( p.pSrcY + 2 * i + j * p.SrcYPitch ) ;
    __m128i s0 = _mm_lddqu_si128 ( pSet ) ;
    __m128i s1 = _mm_srai_epi16 ( s , 2 ) ;
    _mm_storeu_si128 ( pSet , _mm_add_epi8 ( _mm_and_si128 ( _mm_setr_epi16 (
                    (__int16) _mm_extract_epi16 ( s , 1 ) / _mm_extract_epi16 ( s1 , 0 ) ,
                    (__int16) _mm_extract_epi16 ( s , 3 ) / _mm_extract_epi16 ( s1 , 2 ) ,
                    (__int16) _mm_extract_epi16 ( s , 5 ) / _mm_extract_epi16 ( s1 , 4 ) ,
                    (__int16) _mm_extract_epi16 ( s , 7 ) / _mm_extract_epi16 ( s1 , 6 ) ,
                    0 , 0 , 0 , 0 ) , _mm_set1_epi16 ( 0x0FF )) , s0 ) ) ;
}


//==============================================================================
// SetFrame
//==============================================================================
static DWORD WINAPI ThreadFunc_SSE2 ( LPVOID lpParameter )
{
    THREAD_PARAM& p = * (THREAD_PARAM*) lpParameter ;
    const __m128i offsetW = _mm_set1_epi32 ( (__int32) p.pW ) ;
    __m128i wk ;
    __m128i sumcd ;
    for ( int j = p.thr_num ; j < p.Height ; j += p.Threads )
    {
        for ( int i = 0 ; i < p.Width ; i += 4 )
        {
            sumcd = _mm_setzero_si128 () ;
            const int offset = i + p.CacheLOffset + ( j + p.wSpan ) * p.CacheLSpan ;
            const __m128i offsetWv0 = _mm_sub_epi32 ( offsetW , _mm_load_si128 ( (__m128i*) ( p.FrameCache [ 0 ] + offset ) ) ) ;
            const int offsetCache = offset - p.wSpan * p.CacheLSpan - p.wSpan ;
            for ( int dt = -p.tSpan ; dt <= p.tSpan ; dt ++ )
            {
                const __int32 * pv0j = p.FrameCache [ dt ] + offsetCache ;
                for ( int dj = -p.wSpan ; dj <= p.wSpan ; dj ++ , pv0j += p.CacheLSpan )
                {
                    const __int32 * pv0i = pv0j ;
                    for ( int di = -p.wSpan ; di <= p.wSpan ; di ++ , pv0i ++ )
                    {
                        const __m128i AdressW = _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) pv0i ) ) ;
                        CopyMemory ( & wk.m128i_i32 [ 0 ] , (const void*) AdressW.m128i_i32 [ 0 ] , 4 ) ;
                        CopyMemory ( & wk.m128i_i32 [ 1 ] , (const void*) AdressW.m128i_i32 [ 1 ] , 4 ) ;
                        CopyMemory ( & wk.m128i_i32 [ 2 ] , (const void*) AdressW.m128i_i32 [ 2 ] , 4 ) ;
                        CopyMemory ( & wk.m128i_i32 [ 3 ] , (const void*) AdressW.m128i_i32 [ 3 ] , 4 ) ;
                        sumcd = _mm_add_epi32 ( sumcd , wk ) ;
                        sumcd = _mm_add_epi32 ( sumcd , _mm_setr_epi32 (
                            * (const int*) AdressW.m128i_i32 [ 0 ] ,
                            * (const int*) AdressW.m128i_i32 [ 1 ] ,
                            * (const int*) AdressW.m128i_i32 [ 2 ] ,
                            * (const int*) AdressW.m128i_i32 [ 3 ] ) ) ;
                    }
                }
            }
            p.SetFrame ( p , i , j , sumcd ) ;
        }
    }
    return 0 ;
}

static DWORD WINAPI ThreadFunc_SSE2_3x3 ( LPVOID lpParameter )
{
    THREAD_PARAM& p = * (THREAD_PARAM*) lpParameter ;
    const __m128i offsetW = _mm_set1_epi32 ( (__int32) p.pW ) ;
    __m128i AdressW [ 9 ] ;
    __m128i sumcd ;
    for ( int j = p.thr_num ; j < p.Height ; j += p.Threads )
    {
        for ( int i = 0 ; i < p.Width ; i += 4 )
        {
            sumcd = _mm_setzero_si128 () ;
            const int offset = i + p.CacheLOffset + ( j + 1 ) * p.CacheLSpan ;
            const __m128i offsetWv0 = _mm_sub_epi32 ( offsetW , _mm_load_si128 ( (__m128i*) ( p.FrameCache [ 0 ] + offset ) ) ) ;
            for ( int dt = -p.tSpan ; dt <= p.tSpan ; dt ++ )
            {
                const __int32 * pv0j = p.FrameCache [ dt ] + offset ;
                _mm_store_si128 ( AdressW + 0 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0j - 1 - p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 1 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( pv0j + 0 - p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 2 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0j + 1 - p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 3 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0j - 1 + 0            ) ) ) ) ;
                _mm_store_si128 ( AdressW + 4 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( pv0j + 0 + 0            ) ) ) ) ;
                _mm_store_si128 ( AdressW + 5 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0j + 1 + 0            ) ) ) ) ;
                _mm_store_si128 ( AdressW + 6 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0j - 1 + p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 7 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( pv0j + 0 + p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 8 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0j + 1 + p.CacheLSpan ) ) ) ) ;

#define GETADRESS(idx_m) _mm_set_epi32 ( \
        * (const int*) AdressW [ idx_m ].m128i_i32 [ 3 ] , \
        * (const int*) AdressW [ idx_m ].m128i_i32 [ 2 ] , \
        * (const int*) AdressW [ idx_m ].m128i_i32 [ 1 ] , \
        * (const int*) AdressW [ idx_m ].m128i_i32 [ 0 ] )

                sumcd = _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 (
                    GETADRESS ( 0 ) ,
                    GETADRESS ( 1 ) ) ,
                    GETADRESS ( 2 ) ) ,
                    GETADRESS ( 3 ) ) ,
                    GETADRESS ( 4 ) ) ,
                    GETADRESS ( 5 ) ) ,
                    GETADRESS ( 6 ) ) ,
                    GETADRESS ( 7 ) ) ,
                    GETADRESS ( 8 ) ) ,
                    sumcd ) ;
#undef GETADRESS
            }
            p.SetFrame ( p , i , j , sumcd ) ;
        }
    }
    return 0 ;
}

static DWORD WINAPI ThreadFunc_SSE2_5x5 ( LPVOID lpParameter )
{
    THREAD_PARAM& p = * (THREAD_PARAM*) lpParameter ;
    const __m128i offsetW = _mm_set1_epi32 ( (__int32) p.pW ) ;
    __m128i AdressW [ 25 ] ;
    __m128i sumcd ;
    for ( int j = p.thr_num ; j < p.Height ; j += p.Threads )
    {
        for ( int i = 0 ; i < p.Width ; i += 4 )
        {
            sumcd = _mm_setzero_si128 () ;
            const int offset = i + p.CacheLOffset + ( j + 2 ) * p.CacheLSpan ;
            const __m128i offsetWv0 = _mm_sub_epi32 ( offsetW , _mm_load_si128 ( (__m128i*) ( p.FrameCache [ 0 ] + offset ) ) ) ;
            for ( int dt = -p.tSpan ; dt <= p.tSpan ; dt ++ )
            {
                const __int32 * pv0j = p.FrameCache [ dt ] + offset ;
                const __int32 * pv0jw ;
                _mm_store_si128 ( AdressW +  0 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0jw = pv0j - 2 - ( p.CacheLSpan << 1 ) ) ) ) ) ;
                _mm_store_si128 ( AdressW +  1 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  2 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  3 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  4 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  5 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0jw = pv0j - 2 - p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW +  6 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  7 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  8 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  9 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 10 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0jw = pv0j - 2 + 0 ) ) ) ) ;
                _mm_store_si128 ( AdressW + 11 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 12 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 13 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 14 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 15 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0jw = pv0j - 2 + p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 16 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 17 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 18 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 19 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 20 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( pv0jw = pv0j - 2 + ( p.CacheLSpan << 1 ) ) ) ) ) ;
                _mm_store_si128 ( AdressW + 21 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 22 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 23 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 24 , _mm_add_epi32 ( offsetWv0 , _mm_loadu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;

#define GETADRESS(idx_m) _mm_set_epi32 ( \
        * (const int*) AdressW [ idx_m ].m128i_i32 [ 3 ] , \
        * (const int*) AdressW [ idx_m ].m128i_i32 [ 2 ] , \
        * (const int*) AdressW [ idx_m ].m128i_i32 [ 1 ] , \
        * (const int*) AdressW [ idx_m ].m128i_i32 [ 0 ] )

                sumcd = _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 (
                    _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 (
                    _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 (
                    GETADRESS ( 0 ) ,
                    GETADRESS ( 1 ) ) ,
                    GETADRESS ( 2 ) ) ,
                    GETADRESS ( 3 ) ) ,
                    GETADRESS ( 4 ) ) ,
                    GETADRESS ( 5 ) ) ,
                    GETADRESS ( 6 ) ) ,
                    GETADRESS ( 7 ) ) ,
                    GETADRESS ( 8 ) ) ,
                    GETADRESS ( 9 ) ) ,
                    GETADRESS ( 10 ) ) ,
                    GETADRESS ( 11 ) ) ,
                    GETADRESS ( 12 ) ) ,
                    GETADRESS ( 13 ) ) ,
                    GETADRESS ( 14 ) ) ,
                    GETADRESS ( 15 ) ) ,
                    GETADRESS ( 16 ) ) ,
                    GETADRESS ( 17 ) ) ,
                    GETADRESS ( 18 ) ) ,
                    GETADRESS ( 19 ) ) ,
                    GETADRESS ( 20 ) ) ,
                    GETADRESS ( 21 ) ) ,
                    GETADRESS ( 22 ) ) ,
                    GETADRESS ( 23 ) ) ,
                    GETADRESS ( 24 ) ) ,
                    sumcd ) ;
#undef GETADRESS
            }
            p.SetFrame ( p , i , j , sumcd ) ;
        }
    }
    return 0 ;
}

static DWORD WINAPI ThreadFunc_SSE4 ( LPVOID lpParameter )
{
    THREAD_PARAM& p = * (THREAD_PARAM*) lpParameter ;
    const __m128i offsetW = _mm_set1_epi32 ( (__int32) p.pW ) ;
    __m128i sumcd ;
    for ( int j = p.thr_num ; j < p.Height ; j += p.Threads )
    {
        for ( int i = 0 ; i < p.Width ; i += 4 )
        {
            sumcd = _mm_setzero_si128 () ;
            const int offset = i + p.CacheLOffset + ( j + p.wSpan ) * p.CacheLSpan ;
            const __m128i offsetWv0 = _mm_sub_epi32 ( offsetW , _mm_load_si128 ( (__m128i*) ( p.FrameCache [ 0 ] + offset ) ) ) ;
            const int offsetCache = offset - p.wSpan * p.CacheLSpan - p.wSpan ;
            for ( int dt = -p.tSpan ; dt <= p.tSpan ; dt ++ )
            {
                const __int32 * pv0j = p.FrameCache [ dt ] + offsetCache ;
                for ( int dj = -p.wSpan ; dj <= p.wSpan ; dj ++ , pv0j += p.CacheLSpan )
                {
                    const __int32 * pv0i = pv0j ;
                    for ( int di = -p.wSpan ; di <= p.wSpan ; di ++ , pv0i ++ )
                    {
                        const __m128i AdressW = _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) pv0i ) ) ;
                        sumcd = _mm_add_epi32 ( sumcd , _mm_setr_epi32 (
                            * (const int*) _mm_extract_epi32 ( AdressW , 0 ) ,
                            * (const int*) _mm_extract_epi32 ( AdressW , 1 ) ,
                            * (const int*) _mm_extract_epi32 ( AdressW , 2 ) ,
                            * (const int*) _mm_extract_epi32 ( AdressW , 3 ) ) ) ;
                    }
                }
            }
            p.SetFrame ( p , i , j , sumcd ) ;
        }
    }
    return 0 ;
}

static DWORD WINAPI ThreadFunc_SSE4_3x3 ( LPVOID lpParameter )
{
    THREAD_PARAM& p = * (THREAD_PARAM*) lpParameter ;
    const __m128i offsetW = _mm_set1_epi32 ( (__int32) p.pW ) ;
    __m128i AdressW [ 9 ] ;
    __m128i sumcd ;
    for ( int j = p.thr_num ; j < p.Height ; j += p.Threads )
    {
        for ( int i = 0 ; i < p.Width ; i += 4 )
        {
            sumcd = _mm_setzero_si128 () ;
            const int offset = i + p.CacheLOffset + ( j + 1 ) * p.CacheLSpan ;
            const __m128i offsetWv0 = _mm_sub_epi32 ( offsetW , _mm_load_si128 ( (__m128i*) ( p.FrameCache [ 0 ] + offset ) ) ) ;
            for ( int dt = -p.tSpan ; dt <= p.tSpan ; dt ++ )
            {
                const __int32 * pv0j = p.FrameCache [ dt ] + offset ;
                _mm_store_si128 ( AdressW + 0 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0j - 1 - p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 1 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( pv0j + 0 - p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 2 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0j + 1 - p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 3 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0j - 1 + 0            ) ) ) ) ;
                _mm_store_si128 ( AdressW + 4 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( pv0j + 0 + 0            ) ) ) ) ;
                _mm_store_si128 ( AdressW + 5 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0j + 1 + 0            ) ) ) ) ;
                _mm_store_si128 ( AdressW + 6 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0j - 1 + p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 7 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( pv0j + 0 + p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 8 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0j + 1 + p.CacheLSpan ) ) ) ) ;

#define GETADRESS(idx_m) _mm_set_epi32 ( \
        * (const int*) _mm_extract_epi32 ( AdressW [ idx_m ] , 3 ) , \
        * (const int*) _mm_extract_epi32 ( AdressW [ idx_m ] , 2 ) , \
        * (const int*) _mm_extract_epi32 ( AdressW [ idx_m ] , 1 ) , \
        * (const int*) _mm_extract_epi32 ( AdressW [ idx_m ] , 0 ) )

                sumcd = _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 (
                    GETADRESS ( 0 ) ,
                    GETADRESS ( 1 ) ) ,
                    GETADRESS ( 2 ) ) ,
                    GETADRESS ( 3 ) ) ,
                    GETADRESS ( 4 ) ) ,
                    GETADRESS ( 5 ) ) ,
                    GETADRESS ( 6 ) ) ,
                    GETADRESS ( 7 ) ) ,
                    GETADRESS ( 8 ) ) ,
                    sumcd ) ;
#undef GETADRESS
            }
            p.SetFrame ( p , i , j , sumcd ) ;
        }
    }
    return 0 ;
}

static DWORD WINAPI ThreadFunc_SSE4_5x5 ( LPVOID lpParameter )
{
    THREAD_PARAM& p = * (THREAD_PARAM*) lpParameter ;
    const __m128i offsetW = _mm_set1_epi32 ( (__int32) p.pW ) ;
    __m128i AdressW [ 25 ] ;
    __m128i sumcd ;
    for ( int j = p.thr_num ; j < p.Height ; j += p.Threads )
    {
        for ( int i = 0 ; i < p.Width ; i += 4 )
        {
            sumcd = _mm_setzero_si128 () ;
            const int offset = i + p.CacheLOffset + ( j + 2 ) * p.CacheLSpan ;
            const __m128i offsetWv0 = _mm_sub_epi32 ( offsetW , _mm_load_si128 ( (__m128i*) ( p.FrameCache [ 0 ] + offset ) ) ) ;
            for ( int dt = -p.tSpan ; dt <= p.tSpan ; dt ++ )
            {
                const __int32 * pv0j = p.FrameCache [ dt ] + offset ;
                const __int32 * pv0jw ;
                _mm_store_si128 ( AdressW +  0 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0jw = pv0j - 2 - ( p.CacheLSpan << 1 ) ) ) ) ) ;
                _mm_store_si128 ( AdressW +  1 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  2 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  3 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  4 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  5 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0jw = pv0j - 2 - p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW +  6 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  7 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  8 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW +  9 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 10 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0jw = pv0j - 2 + 0 ) ) ) ) ;
                _mm_store_si128 ( AdressW + 11 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 12 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 13 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 14 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 15 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0jw = pv0j - 2 + p.CacheLSpan ) ) ) ) ;
                _mm_store_si128 ( AdressW + 16 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 17 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 18 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 19 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 20 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( pv0jw = pv0j - 2 + ( p.CacheLSpan << 1 ) ) ) ) ) ;
                _mm_store_si128 ( AdressW + 21 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 22 , _mm_add_epi32 ( offsetWv0 , _mm_load_si128  ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 23 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;
                _mm_store_si128 ( AdressW + 24 , _mm_add_epi32 ( offsetWv0 , _mm_lddqu_si128 ( (const __m128i*) ( ++ pv0jw ) ) ) ) ;

#define GETADRESS(idx_m) _mm_set_epi32 ( \
        * (const int*) _mm_extract_epi32 ( AdressW [ idx_m ] , 3 ) , \
        * (const int*) _mm_extract_epi32 ( AdressW [ idx_m ] , 2 ) , \
        * (const int*) _mm_extract_epi32 ( AdressW [ idx_m ] , 1 ) , \
        * (const int*) _mm_extract_epi32 ( AdressW [ idx_m ] , 0 ) )

                sumcd = _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 (
                    _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 (
                    _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 ( _mm_add_epi32 (
                    GETADRESS ( 0 ) ,
                    GETADRESS ( 1 ) ) ,
                    GETADRESS ( 2 ) ) ,
                    GETADRESS ( 3 ) ) ,
                    GETADRESS ( 4 ) ) ,
                    GETADRESS ( 5 ) ) ,
                    GETADRESS ( 6 ) ) ,
                    GETADRESS ( 7 ) ) ,
                    GETADRESS ( 8 ) ) ,
                    GETADRESS ( 9 ) ) ,
                    GETADRESS ( 10 ) ) ,
                    GETADRESS ( 11 ) ) ,
                    GETADRESS ( 12 ) ) ,
                    GETADRESS ( 13 ) ) ,
                    GETADRESS ( 14 ) ) ,
                    GETADRESS ( 15 ) ) ,
                    GETADRESS ( 16 ) ) ,
                    GETADRESS ( 17 ) ) ,
                    GETADRESS ( 18 ) ) ,
                    GETADRESS ( 19 ) ) ,
                    GETADRESS ( 20 ) ) ,
                    GETADRESS ( 21 ) ) ,
                    GETADRESS ( 22 ) ) ,
                    GETADRESS ( 23 ) ) ,
                    GETADRESS ( 24 ) ) ,
                    sumcd ) ;
#undef GETADRESS
            }
            p.SetFrame ( p , i , j , sumcd ) ;
        }
    }
    return 0 ;
}