// $Date:: 2010-02-06 15:52:49 +0900#$
// $Rev: 11 $

#include "PNLM2.h"
#include <math.h>
#include <emmintrin.h>
#include <smmintrin.h>

//#define PRINTTIME
#ifdef PRINTTIME
#include <stdio.h>
#include <mmsystem.h>
#pragma comment ( lib , "winmm.lib" )
static __int32 tcount = 0 ;
#endif

static const __m128i Low16Mask = _mm_set1_epi16 ( 0x00FF ) ;

static DWORD WINAPI ThreadFunc ( LPVOID lpParameter ) ;
static void SetCacheFrame_YV12_SSE2 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM2& p ) ;
static void SetCacheFrame_YV12_SSE4 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM2& p ) ;
static void SetCacheFrame_YUY2_SSE2 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM2& p ) ;
static void SetCacheFrame_YUY2_SSE4 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM2& p ) ;
static __inline void SetAngleValue_SSE2 ( int cache_idx , THREAD_PARAM2& p ) ;
static __inline void SetAngleValue_SSE4 ( int cache_idx , THREAD_PARAM2& p ) ;
static void SetFrame_YV12_SSE2 ( __m128i& sumY ,  __m128i& sumC , __m128i * d ) ;
static void SetFrame_YV12_SSE4 ( __m128i& sumY ,  __m128i& sumC , __m128i * d ) ;
static void SetFrame_YUY2_SSE2 ( __m128i& sumY ,  __m128i& sumC , __m128i * d ) ;
static void SetFrame_YUY2_SSE4 ( __m128i& sumY ,  __m128i& sumC , __m128i * d ) ;
static DWORD WINAPI ThreadFunc_SSE2 ( LPVOID lpParameter ) ;
static DWORD WINAPI ThreadFunc_SSE4 ( LPVOID lpParameter ) ;

/**************************************************************************************************/
/* PNLM2                                                                                      */
/**************************************************************************************************/
//==============================================================================
// constructor
//==============================================================================
PNLM2::PNLM2(PClip _child, int strength, int wspan, int tspan, int selfweight, int angleweight, int threads, int opt, IScriptEnvironment* env) :
    GenericVideoFilter(_child) ,
    Height      ( this->vi.height   ) ,
    Width       ( this->vi.width    ) ,
    Strength    ( strength          ) , // ڂWB0`22 ͈̔͂ŏB
    SelfWeight  ( selfweight        ) , // g̏d݌WB-9`20 ͈̔͂ŏB
    AngleWeight ( 12 - angleweight  ) , // Xd݌WB0`12 ͈̔͂ŏBrbgVtgB
    wSpan       ( wspan             ) , // ͈́B0` ͈̔͂ŏB1Ȃ3x3A2Ȃ5x5̃}gNXB
    tSpan       ( tspan             ) , // ͈́B0` ͈̔͂ŏB1ȂOPt[̎ԎB
    Threads     ( threads           )   // XbhB
{
	if( strength < 0 || 110 < strength )
		env->ThrowError("PNLM2:'STRENGTH' must be 0 to 110.");

    if ( selfweight < -9 || selfweight > 20 )
        env->ThrowError("PNLM2:'SELFWEIGHT' must be -9 to 20.");

	if( angleweight < 0 || 12 < angleweight )
		env->ThrowError("PNLM2:'ANGLEWEIGHT' must be 0 to 12.");

	if( wspan < 0 )
		env->ThrowError("PNLM2:'WSPAN' must be greater than 0.");

	if( tspan < 0 )
		env->ThrowError("PNLM2:'TSPAN' must be greater than 0.");

	if( threads < 1 || 15 < threads )
		env->ThrowError("PNLM2:'THREADS' must be 1 to 15.");

	if( opt < 0 || 2 < opt )
		env->ThrowError("PNLM2:'OPT' must be 0 to 2.");

    if( opt == 0 )
    {
	    unsigned int output_ecx;
	    unsigned int output_edx;
	    __asm{
		    push eax
		    push ebx
		    push ecx
		    push edx

		    mov eax, 1

		    CPUID

		    mov output_ecx, edx
		    mov output_edx, edx

		    pop edx
		    pop ecx
		    pop ebx
		    pop eax
	    }

	    if ( ! ( output_edx & 0x04000000 ) )
		    env->ThrowError("PNLM2:this machine do not support SSE2.");

        opt = ( output_ecx & 0x00080000 ) ? 2 : 1 ;
    }
    
    const int area = ( 2 * wspan + 1 ) * ( 2 * wspan + 1 ) * ( 2 * tspan + 1 ) ;
    if ( area > 81 )
        env->ThrowError("PNLM2:'SELFWEIGHT' must be (2*tspan+1)*(2*wspan+1)^2 <= 81 .");

    int w = ( 256 - selfweight ) / area ;
    if ( w >= 16 ) w = 15 ;
    if ( w + selfweight <= 0 )
        env->ThrowError("PNLM2:'SELFWEIGHT' is too small.");

    this->pW = (unsigned __int8 *) _aligned_malloc ( 256 , 16 ) ;
    const double H = - pow ( 10, (double) strength / -22.0) ;
    for ( int i = 0 ; i < 255 ; i ++ )
        this->pW [ i ] = (unsigned __int8) ( exp ( H * i * i ) * w + 0.499 );
    this->pW [ 255 ] = 0 ;

    // 镝i͈͂ 16 ̔{ɂj
    this->SrcWidth = this->Width & ~ ( this->vi.IsYV12 () ? 0x0F : 0x07 ) ;

    // 炵
    this->CacheLOffset = 8 ;

    // PC̃oCg
    this->CacheLSpan = this->SrcWidth + this->CacheLOffset * 2 ;

    // LbVobt@
    this->Y_FrameCache = ( (__int16**) malloc ( ( 2 * this->tSpan + 1 ) * sizeof (__int16*) ) ) + this->tSpan ;
    this->V_FrameCache = ( (__int16**) malloc ( ( 2 * this->tSpan + 1 ) * sizeof (__int16*) ) ) + this->tSpan ;
    this->H_FrameCache = ( (__int16**) malloc ( ( 2 * this->tSpan + 1 ) * sizeof (__int16*) ) ) + this->tSpan ;
    for ( int i = -this->tSpan ; i <= this->tSpan  ; i ++ )
    {
        this->Y_FrameCache [ i ] = (__int16*) _aligned_malloc ( this->CacheLSpan * ( this->Height + 2 * max ( this->wSpan , 1 ) ) * sizeof (__int16) , 16 ) ;
        this->V_FrameCache [ i ] = (__int16*) _aligned_malloc ( this->CacheLSpan * ( this->Height + 2 * max ( this->wSpan , 1 ) ) * sizeof (__int16) , 16 ) ;
        this->H_FrameCache [ i ] = (__int16*) _aligned_malloc ( this->CacheLSpan * ( this->Height + 2 * max ( this->wSpan , 1 ) ) * sizeof (__int16) , 16 ) ;
    }
    this->AvgWork = (__int16*) _aligned_malloc ( this->CacheLSpan * ( this->Height + 2 * max ( this->wSpan , 1 ) ) * sizeof (__int16) , 16 ) ;

    this->PreFrameNum = -2 ;

    this->ThreadFunc = ( opt == 1 ) ? ::ThreadFunc_SSE2 : ::ThreadFunc_SSE4 ;
    this->SetCacheFrame = ( opt == 1 ) ? ( this->vi.IsYV12 () ? ::SetCacheFrame_YV12_SSE2 : ::SetCacheFrame_YUY2_SSE2 )
                                       : ( this->vi.IsYV12 () ? ::SetCacheFrame_YV12_SSE4 : ::SetCacheFrame_YUY2_SSE4 ) ;
    this->SetFrame = ( opt == 1 ) ? ( this->vi.IsYV12 () ? ::SetFrame_YV12_SSE2 : ::SetFrame_YUY2_SSE2 )
                                  : ( this->vi.IsYV12 () ? ::SetFrame_YV12_SSE4 : ::SetFrame_YUY2_SSE4 ) ;
}

//==============================================================================
// destructor
//==============================================================================
PNLM2::~PNLM2()
{
    for ( int i = -this->tSpan ; i <= this->tSpan  ; i ++ )
    {
        if ( this->Y_FrameCache && this->Y_FrameCache [ i ] ) _aligned_free ( this->Y_FrameCache [ i ] ) ;
        if ( this->V_FrameCache && this->V_FrameCache [ i ] ) _aligned_free ( this->V_FrameCache [ i ] ) ;
        if ( this->H_FrameCache && this->H_FrameCache [ i ] ) _aligned_free ( this->H_FrameCache [ i ] ) ;
    }
    if ( this->Y_FrameCache ) free ( this->Y_FrameCache - this->tSpan ) ;
    if ( this->V_FrameCache ) free ( this->V_FrameCache - this->tSpan ) ;
    if ( this->H_FrameCache ) free ( this->H_FrameCache - this->tSpan ) ;
    if ( this->AvgWork ) _aligned_free ( this->AvgWork ) ;
    if ( this->pW      ) _aligned_free ( this->pW      ) ;

#ifdef PRINTTIME
    char strtime [ 256 ] ;
    _snprintf ( strtime , sizeof (strtime) , "%f\r\n" , ( (float) tcount ) / 1000 ) ;
    OutputDebugString ( strtime ) ;
#endif // PRINTTIME
}

//==============================================================================
// GetFrame
//==============================================================================
PVideoFrame PNLM2::GetFrame (int n, IScriptEnvironment* env)
{
    PVideoFrame src = this->child->GetFrame ( n , env ) ;
    env->MakeWritable ( & src ) ;

    BYTE* pSrcY = src->GetWritePtr () ;

#ifdef PRINTTIME
    tcount -= timeGetTime () ;
#endif // PRINTTIME

    if ( this->PreFrameNum == -2 )
    {
        // ߂ GetFrame Ă΂ꂽƂ
        this->SrcYPitch = src->GetPitch() ;
    }

    THREAD_PARAM2 p (
        this->pW ,
        this->wSpan ,
        this->tSpan ,
        this->pW [ 0 ] + this->SelfWeight ,
        this->AngleWeight ,
        this->Threads ,
        this->SrcYPitch ,
        this->Height ,
        this->SrcWidth ,
        this->CacheLSpan ,
        this->CacheLOffset ,
        this->vi.IsYV12 () ,
        this->Y_FrameCache ,
        this->V_FrameCache ,
        this->H_FrameCache ,
        this->AvgWork ,
        pSrcY ,
        this->Threads ,
        this->SetFrame
        ) ;
    
    if ( this->PreFrameNum + 1 != n )
    {
        for ( int i = -this->tSpan ; i < this->tSpan  ; i ++ )
        {
            if ( n + i < 0 )
            {
                this->SetCacheFrame ( i , 0 , env , this->child , p ) ;
            }
            else if ( n + i >= this->vi.num_frames )
            {
                this->SetCacheFrame ( i , this->vi.num_frames - 1 , env , this->child , p ) ;
            }
            else
            {
                this->SetCacheFrame ( i , n + i , env , this->child , p ) ;
            }
        }
    }
    else
    {
        __int16* pTmp ;
        pTmp = this->Y_FrameCache [ -this->tSpan ] ;
        memcpy ( this->Y_FrameCache - this->tSpan , this->Y_FrameCache - this->tSpan + 1 ,
            2 * this->tSpan * sizeof (BYTE*) ) ;
        this->Y_FrameCache [ this->tSpan ] = pTmp ;
        
        pTmp = this->V_FrameCache [ -this->tSpan ] ;
        memcpy ( this->V_FrameCache - this->tSpan , this->V_FrameCache - this->tSpan + 1 ,
            2 * this->tSpan * sizeof (BYTE*) ) ;
        this->V_FrameCache [ this->tSpan ] = pTmp ;
        
        pTmp = this->H_FrameCache [ -this->tSpan ] ;
        memcpy ( this->H_FrameCache - this->tSpan , this->H_FrameCache - this->tSpan + 1 ,
            2 * this->tSpan * sizeof (BYTE*) ) ;
        this->H_FrameCache [ this->tSpan ] = pTmp ;
    }
    this->PreFrameNum = n ;

    if ( n + this->tSpan >= this->vi.num_frames )
    {
        this->SetCacheFrame ( this->tSpan , this->vi.num_frames - 1 , env , this->child , p ) ;
    }
    else
    {
        this->SetCacheFrame ( this->tSpan , n + this->tSpan , env , this->child , p ) ;
    }

#if 0
    this->ThreadFunc ( & p ) ;
#else
    HANDLE hThr [ 15 ] ;
    THREAD_PARAM2 * sParam = (THREAD_PARAM2*) alloca ( this->Threads * sizeof (THREAD_PARAM2) ) ;
    for ( int i =0 ; i < this->Threads ; i ++ )
    {
        memcpy ( sParam + i , & p , sizeof (THREAD_PARAM2) ) ;
        sParam [ i ].thr_num = i ;
    }
    for ( int i = 0 ; i < this->Threads ; i ++ )
        hThr [ i ] = ::CreateThread ( NULL , 0 , this->ThreadFunc , sParam + i , 0 , NULL ) ;
    for ( int i = 0 ; i < this->Threads ; i ++ )
    {
        ::WaitForSingleObject ( hThr [ i ] , INFINITE ) ;
        ::CloseHandle ( hThr [ i ] ) ;
    }
#endif

#ifdef PRINTTIME
    tcount += timeGetTime () ;
#endif // PRINTTIME
    return src ;
}

//==============================================================================
// SetCacheFrame
//==============================================================================
static void SetCacheFrame_YV12_SSE2 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM2& p )
{
    __int16 * ppp = p.Y_FrameCache [ cache_idx ] + max ( p.wSpan , 1 ) * p.CacheLSpan ;
    __int16 * qqq ;
    const BYTE * src = child->GetFrame ( frame_num , env )->GetReadPtr () ;
    for ( int j = 0 ; j < p.Height ; j ++ )
    {
        const BYTE * sss = src + j * p.SrcYPitch ;

        // [
        _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi16 ( * sss ) ) ;
        ppp += 8 ;

        // g
        for ( int i = 0 ; i < p.Width ; i += 16 , ppp += 16 , sss += 16 )
        {
            const __m128i wk = _mm_loadu_si128 ( (__m128i*) sss ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 0  ) , _mm_unpacklo_epi8 ( wk , _mm_setzero_si128 () ) ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 8  ) , _mm_unpackhi_epi8 ( wk , _mm_setzero_si128 () ) ) ;
        }

        // E[
        _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi16 ( sss [ -1 ] ) ) ;
        ppp += 8 ;
    }

    // [
    qqq = p.Y_FrameCache [ cache_idx ] + ( p.Height + max ( p.wSpan , 1 ) - 1 ) * p.CacheLSpan ;
    ppp = p.Y_FrameCache [ cache_idx ] + ( p.Height + max ( p.wSpan , 1 ) ) * p.CacheLSpan ;
    for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int16) ) ;

    // [
    qqq = p.Y_FrameCache [ cache_idx ] + max ( p.wSpan , 1 ) * p.CacheLSpan ;
    ppp = p.Y_FrameCache [ cache_idx ] ;
    for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int16) ) ;

    ::SetAngleValue_SSE2 ( cache_idx , p ) ;
}

static void SetCacheFrame_YV12_SSE4 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM2& p )
{
    __int16 * ppp = p.Y_FrameCache [ cache_idx ] + max ( p.wSpan , 1 ) * p.CacheLSpan ;
    __int16 * qqq ;
    const BYTE * src = child->GetFrame ( frame_num , env )->GetReadPtr () ;
    for ( int j = 0 ; j < p.Height ; j ++ )
    {
        const BYTE * sss = src + j * p.SrcYPitch ;

        // [
        _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi16 ( * sss ) ) ;
        ppp += 8 ;

        // g
        for ( int i = 0 ; i < p.Width ; i += 16 , ppp += 16 , sss += 16 )
        {
            const __m128i wk = _mm_lddqu_si128 ( (__m128i*) sss ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 0  ) , _mm_cvtepu8_epi16 ( wk                         ) ) ;
            _mm_store_si128 ( (__m128i*) ( ppp + 8  ) , _mm_cvtepu8_epi16 ( _mm_srli_si128 ( wk , 8  ) ) ) ;
        }

        // E[
        _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi16 ( sss [ -1 ] ) ) ;
        ppp += 8 ;
    }

    // [
    qqq = p.Y_FrameCache [ cache_idx ] + ( p.Height + max ( p.wSpan , 1 ) - 1 ) * p.CacheLSpan ;
    ppp = p.Y_FrameCache [ cache_idx ] + ( p.Height + max ( p.wSpan , 1 ) ) * p.CacheLSpan ;
    for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int16) ) ;

    // [
    qqq = p.Y_FrameCache [ cache_idx ] + max ( p.wSpan , 1 ) * p.CacheLSpan ;
    ppp = p.Y_FrameCache [ cache_idx ] ;
    for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int16) ) ;

    ::SetAngleValue_SSE4 ( cache_idx , p ) ;
}

static void SetCacheFrame_YUY2_SSE2 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM2& p )
{
    __int16 * ppp = p.Y_FrameCache [ cache_idx ] + max ( p.wSpan , 1 ) * p.CacheLSpan ;
    __int16 * qqq ;
    const BYTE * src = child->GetFrame ( frame_num , env )->GetReadPtr () ;
    static const __m128i mask = _mm_set1_epi16 ( 0x0FF ) ;
    for ( int j = 0 ; j < p.Height ; j ++ )
    {
        const BYTE * sss = src + j * p.SrcYPitch ;

        // [
        _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi16 ( * sss ) ) ;
        ppp += 8 ;

        // gi16 oCgŏł镪j
        for ( int i = 0 ; i < ( p.Width & ~ 0x07 ) ; i += 8 , ppp += 8 , sss += 16 )
            _mm_store_si128 ( (__m128i*) ppp , _mm_and_si128 ( _mm_loadu_si128 ( (__m128i*) sss ) , mask ) ) ;

        // E[
        _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi16 ( sss [ -1 ] ) ) ;
        ppp += 8 ;
    }

    // [
    qqq = p.Y_FrameCache [ cache_idx ] + ( p.Height + max ( p.wSpan , 1 ) - 1 ) * p.CacheLSpan ;
    ppp = p.Y_FrameCache [ cache_idx ] + ( p.Height + max ( p.wSpan , 1 ) ) * p.CacheLSpan ;
    for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int16) ) ;

    // [
    qqq = p.Y_FrameCache [ cache_idx ] + max ( p.wSpan , 1 ) * p.CacheLSpan ;
    ppp = p.Y_FrameCache [ cache_idx ] ;
    for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int16) ) ;

    ::SetAngleValue_SSE2 ( cache_idx , p ) ;
}

static void SetCacheFrame_YUY2_SSE4 ( int cache_idx , int frame_num , IScriptEnvironment* env , PClip child , THREAD_PARAM2& p )
{
    __int16 * ppp = p.Y_FrameCache [ cache_idx ] + max ( p.wSpan , 1 ) * p.CacheLSpan ;
    __int16 * qqq ;
    const BYTE * src = child->GetFrame ( frame_num , env )->GetReadPtr () ;
    static const __m128i mask = _mm_set1_epi16 ( 0x0FF ) ;
    for ( int j = 0 ; j < p.Height ; j ++ )
    {
        const BYTE * sss = src + j * p.SrcYPitch ;

        // [
        _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi16 ( * sss ) ) ;
        ppp += 8 ;

        // gi16 oCgŏł镪j
        for ( int i = 0 ; i < ( p.Width & ~ 0x07 ) ; i += 8 , ppp += 8 , sss += 16 )
            _mm_store_si128 ( (__m128i*) ppp , _mm_and_si128 ( _mm_lddqu_si128 ( (__m128i*) sss ) , mask ) ) ;

        // E[
        _mm_store_si128 ( (__m128i*) ppp , _mm_set1_epi16 ( sss [ -1 ] ) ) ;
        ppp += 8 ;
    }

    // [
    qqq = p.Y_FrameCache [ cache_idx ] + ( p.Height + max ( p.wSpan , 1 ) - 1 ) * p.CacheLSpan ;
    ppp = p.Y_FrameCache [ cache_idx ] + ( p.Height + max ( p.wSpan , 1 ) ) * p.CacheLSpan ;
    for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int16) ) ;

    // [
    qqq = p.Y_FrameCache [ cache_idx ] + max ( p.wSpan , 1 ) * p.CacheLSpan ;
    ppp = p.Y_FrameCache [ cache_idx ] ;
    for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , ppp += p.CacheLSpan )
        CopyMemory ( ppp , qqq , p.CacheLSpan * sizeof(__int16) ) ;

    ::SetAngleValue_SSE4 ( cache_idx , p ) ;
}

static __inline void SetAngleValue_SSE2 ( int cache_idx , THREAD_PARAM2& p )
{
    __m128i * Y = (__m128i *) p.Y_FrameCache [ cache_idx ] ;
    __m128i * V = (__m128i *) p.V_FrameCache [ cache_idx ] ;
    __m128i * H = (__m128i *) p.H_FrameCache [ cache_idx ] ;
    __m128i * A = (__m128i *) p.AvgWork;
    {
        // PxώZo
        const __m128i * s0 = (const __m128i *) ( Y - 1 + max ( p.wSpan , 1 ) * p.CacheLSpan + p.CacheLOffset ) ;
        const __m128i * s1 = (const __m128i *) ( Y     + max ( p.wSpan , 1 ) * p.CacheLSpan + p.CacheLOffset ) ;
        const __m128i * s2 = (const __m128i *) ( Y + 1 + max ( p.wSpan , 1 ) * p.CacheLSpan + p.CacheLOffset ) ;
        __m128i * d = (__m128i *) ( V + max ( p.wSpan , 1 ) * p.CacheLSpan ) ;
        for ( int i = 0 ; i < p.Height ; i ++ )
        {
            // [
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_add_epi16 ( _mm_loadu_si128 ( s0 + 1 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 + 1 ) , _mm_loadu_si128 ( s2 + 1 ) ) ).m128i_i16 [ 0 ] ) ) ;
            d ++ ; s0 ++ ; s1 ++ ; s2 ++ ;

            // 
            for ( int j = 8 ; j < ( p.CacheLSpan >> 3 ) - 8 ; j ++ , d ++ , s0 ++ , s1 ++ , s2 ++ )
                _mm_store_si128 ( d , _mm_add_epi16 ( _mm_loadu_si128 ( s0 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 ) , _mm_loadu_si128 ( s2 ) ) ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( d [ -1 ].m128i_i16 [ 7 ] ) ) ;
            d ++ ; s0 ++ ; s1 ++ ; s2 ++ ;
        }
        // [
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d - ( p.CacheLSpan >> 3 ) , d , p.CacheLSpan * sizeof (__int16) ) ;

        // [
        d = (__m128i *) V  ;
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d , d + ( p.CacheLSpan >> 3 ) , p.CacheLSpan * sizeof (__int16) ) ;

        d = (__m128i *) ( A + max ( p.wSpan , 1 ) * p.CacheLSpan ) ;
        s0 = (const __m128i *) ( V + ( max ( p.wSpan , 1 ) - 1 ) * p.CacheLSpan ) ;
        s1 = s0 + p.CacheLSpan ;
        s2 = s1 + p.CacheLSpan ;
        for ( int i = 0 ; i < p.Height ; i ++ )
        {
            // [
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_add_epi16 ( _mm_load_si128 ( s0 + 1 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 + 1 ) , _mm_load_si128 ( s2 + 1 ) ) ).m128i_i16 [ 0 ] ) ) ;
            d ++ ; s0 ++ ; s1 ++ ; s2 ++ ;

            // 
            for ( int j = 8 ; j < ( p.CacheLSpan >> 3 ) - 8 ; j ++ , d ++ , s0 ++ , s1 ++ , s2 ++ )
                _mm_store_si128 ( d , _mm_add_epi16 ( _mm_load_si128 ( s0 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 ) , _mm_load_si128 ( s2 ) ) ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( d [ -1 ].m128i_i16 [ 7 ] ) ) ;
            d ++ ; s0 ++ ; s1 ++ ; s2 ++ ;
        }
        // [
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d - ( p.CacheLSpan >> 3 ) , d , p.CacheLSpan * sizeof (__int16) ) ;

        // [
        d = (__m128i *) A  ;
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d , d + ( p.CacheLSpan >> 3 ) , p.CacheLSpan * sizeof (__int16) ) ;
    }

    {
        // PxXZo
        __m128i * d = (__m128i *) ( H + max ( p.wSpan , 1 ) * p.CacheLSpan ) ;
        const __m128i * s0 = (const __m128i *) ( A  - 1 + max ( p.wSpan , 1 ) * p.CacheLSpan) ;
        const __m128i * s1 = (const __m128i *) ( A  + 1 + max ( p.wSpan , 1 ) * p.CacheLSpan) ;
        for ( int i = 0 ; i < p.Height ; i ++ )
        {
            // [
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_sub_epi16 ( _mm_loadu_si128 ( s0 + 1 ) , _mm_loadu_si128 ( s1 + 1 ) ).m128i_i16 [ 0 ] ) ) ;
            d ++ ; s0 ++ ; s1 ++ ;

            // 
            for ( int j = 8 ; j < ( p.Width >> 3 ) - 8 ; j ++ , d ++ , s0 ++ , s1 ++ )
                _mm_store_si128 ( d , _mm_sub_epi16 ( _mm_loadu_si128 ( s0 ) , _mm_loadu_si128 ( s1 ) ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( d [ -1 ].m128i_i16 [ 7 ] ) ) ;
            d ++ ; s0 ++ ; s1 ++ ;
        }
        // [
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d - ( p.CacheLSpan >> 3 ) , d , p.CacheLSpan * sizeof (__int16) ) ;

        // [
        d = (__m128i *) H  ;
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d , d + ( p.CacheLSpan >> 3 ) , p.CacheLSpan * sizeof (__int16) ) ;
    }

    {
        // PxXZo
        __m128i * d = (__m128i *) ( V + max ( p.wSpan , 1 ) * p.CacheLSpan ) ;
        const __m128i * s0 = (const __m128i *) ( A + ( max ( p.wSpan , 1 ) - 1 ) * p.CacheLSpan) ;
        const __m128i * s1 = (const __m128i *) ( A + ( max ( p.wSpan , 1 ) + 1 ) * p.CacheLSpan) ;
        for ( int i = 0 ; i < p.Height ; i ++ )
        {
            // [
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_sub_epi16 ( _mm_loadu_si128 ( s0 + 1 ) , _mm_loadu_si128 ( s1 + 1 ) ).m128i_i16 [ 0 ] ) ) ;
            d ++ ; s0 ++ ; s1 ++ ;

            // 
            for ( int j = 8 ; j < ( p.Width >> 3 ) - 8 ; j ++ , d ++ , s0 ++ , s1 ++ )
                _mm_store_si128 ( d , _mm_sub_epi16 ( _mm_loadu_si128 ( s0 ) , _mm_loadu_si128 ( s1 ) ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( d [ -1 ].m128i_i16 [ 7 ] ) ) ;
            d ++ ; s0 ++ ; s1 ++ ;
        }
        // [
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d - ( p.CacheLSpan >> 3 ) , d , p.CacheLSpan * sizeof (__int16) ) ;

        // [
        d = (__m128i *) V ;
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d , d + ( p.CacheLSpan >> 3 ) , p.CacheLSpan * sizeof (__int16) ) ;
    }
}

static __inline void SetAngleValue_SSE4 ( int cache_idx , THREAD_PARAM2& p )
{
    __m128i * Y = (__m128i *) p.Y_FrameCache [ cache_idx ] ;
    __m128i * V = (__m128i *) p.V_FrameCache [ cache_idx ] ;
    __m128i * H = (__m128i *) p.H_FrameCache [ cache_idx ] ;
    __m128i * A = (__m128i *) p.AvgWork;
    {
        // PxώZo
        const __m128i * s0 = (const __m128i *) ( Y - 1 + max ( p.wSpan , 1 ) * p.CacheLSpan + p.CacheLOffset ) ;
        const __m128i * s1 = (const __m128i *) ( Y     + max ( p.wSpan , 1 ) * p.CacheLSpan + p.CacheLOffset ) ;
        const __m128i * s2 = (const __m128i *) ( Y + 1 + max ( p.wSpan , 1 ) * p.CacheLSpan + p.CacheLOffset ) ;
        __m128i * d = (__m128i *) ( V + max ( p.wSpan , 1 ) * p.CacheLSpan ) ;
        for ( int i = 0 ; i < p.Height ; i ++ )
        {
            // [
            _mm_store_si128 ( d , _mm_add_epi16 ( _mm_lddqu_si128 ( s0 + 1 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 + 1 ) , _mm_lddqu_si128 ( s2 + 1 ) ) ) ) ;
            d ++ ; s0 ++ ; s1 ++ ; s2 ++ ;

            // 
            for ( int j = 8 ; j < ( p.CacheLSpan >> 3 ) - 8 ; j ++ , d ++ , s0 ++ , s1 ++ , s2 ++ )
                _mm_store_si128 ( d , _mm_add_epi16 ( _mm_lddqu_si128 ( s0 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 ) , _mm_lddqu_si128 ( s2 ) ) ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
            d ++ ; s0 ++ ; s1 ++ ; s2 ++ ;
        }
        // [
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d - ( p.CacheLSpan >> 3 ) , d , p.CacheLSpan * sizeof (__int16) ) ;

        // [
        d = (__m128i *) V  ;
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d , d + ( p.CacheLSpan >> 3 ) , p.CacheLSpan * sizeof (__int16) ) ;

        d = (__m128i *) ( A + max ( p.wSpan , 1 ) * p.CacheLSpan ) ;
        s0 = (const __m128i *) ( V + ( max ( p.wSpan , 1 ) - 1 ) * p.CacheLSpan ) ;
        s1 = s0 + p.CacheLSpan ;
        s2 = s1 + p.CacheLSpan ;
        for ( int i = 0 ; i < p.Height ; i ++ )
        {
            // [
            _mm_store_si128 ( d , _mm_add_epi16 ( _mm_load_si128 ( s0 + 1 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 + 1 ) , _mm_load_si128 ( s2 + 1 ) ) ) ) ;
            d ++ ; s0 ++ ; s1 ++ ; s2 ++ ;

            // 
            for ( int j = 8 ; j < ( p.CacheLSpan >> 3 ) - 8 ; j ++ , d ++ , s0 ++ , s1 ++ , s2 ++ )
                _mm_store_si128 ( d , _mm_add_epi16 ( _mm_load_si128 ( s0 ) , _mm_add_epi16 ( _mm_load_si128 ( s1 ) , _mm_load_si128 ( s2 ) ) ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
            d ++ ; s0 ++ ; s1 ++ ; s2 ++ ;
        }
        // [
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d - ( p.CacheLSpan >> 3 ) , d , p.CacheLSpan * sizeof (__int16) ) ;

        // [
        d = (__m128i *) A  ;
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d , d + ( p.CacheLSpan >> 3 ) , p.CacheLSpan * sizeof (__int16) ) ;
    }

    {
        // PxXZo
        __m128i * d = (__m128i *) ( H + max ( p.wSpan , 1 ) * p.CacheLSpan ) ;
        const __m128i * s0 = (const __m128i *) ( A  - 1 + max ( p.wSpan , 1 ) * p.CacheLSpan) ;
        const __m128i * s1 = (const __m128i *) ( A  + 1 + max ( p.wSpan , 1 ) * p.CacheLSpan) ;
        for ( int i = 0 ; i < p.Height ; i ++ )
        {
            // [
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( _mm_sub_epi16 ( _mm_lddqu_si128 ( s0 + 1 ) , _mm_lddqu_si128 ( s1 + 1 ) ) , 0 ) ) ) ;
            d ++ ; s0 ++ ; s1 ++ ;

            // 
            for ( int j = 8 ; j < ( p.Width >> 3 ) - 8 ; j ++ , d ++ , s0 ++ , s1 ++ )
                _mm_store_si128 ( d , _mm_sub_epi16 ( _mm_lddqu_si128 ( s0 ) , _mm_lddqu_si128 ( s1 ) ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
            d ++ ; s0 ++ ; s1 ++ ;
        }
        // [
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d - ( p.CacheLSpan >> 3 ) , d , p.CacheLSpan * sizeof (__int16) ) ;

        // [
        d = (__m128i *) H  ;
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d , d + ( p.CacheLSpan >> 3 ) , p.CacheLSpan * sizeof (__int16) ) ;
    }

    {
        // PxXZo
        __m128i * d = (__m128i *) ( V + max ( p.wSpan , 1 ) * p.CacheLSpan ) ;
        const __m128i * s0 = (const __m128i *) ( A + ( max ( p.wSpan , 1 ) - 1 ) * p.CacheLSpan) ;
        const __m128i * s1 = (const __m128i *) ( A + ( max ( p.wSpan , 1 ) + 1 ) * p.CacheLSpan) ;
        for ( int i = 0 ; i < p.Height ; i ++ )
        {
            // [
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( _mm_sub_epi16 ( _mm_lddqu_si128 ( s0 + 1 ) , _mm_lddqu_si128 ( s1 + 1 ) ) , 0 ) ) ) ;
            d ++ ; s0 ++ ; s1 ++ ;

            // 
            for ( int j = 8 ; j < ( p.Width >> 3 ) - 8 ; j ++ , d ++ , s0 ++ , s1 ++ )
                _mm_store_si128 ( d , _mm_sub_epi16 ( _mm_lddqu_si128 ( s0 ) , _mm_lddqu_si128 ( s1 ) ) ) ;

            // E[
            _mm_store_si128 ( d , _mm_set1_epi16 ( _mm_extract_epi16 ( d [ -1 ] , 7 ) ) ) ;
            d ++ ; s0 ++ ; s1 ++ ;
        }
        // [
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d - ( p.CacheLSpan >> 3 ) , d , p.CacheLSpan * sizeof (__int16) ) ;

        // [
        d = (__m128i *) V ;
        for ( int i = 0 ; i < max ( p.wSpan , 1 ) ; i ++ , d+= ( p.CacheLSpan >> 3 ) )
            CopyMemory ( d , d + ( p.CacheLSpan >> 3 ) , p.CacheLSpan * sizeof (__int16) ) ;
    }
}


//==============================================================================
// SetFrame
//==============================================================================
static void SetFrame_YV12_SSE2 ( __m128i& sumY , __m128i& sumC , __m128i * d )
{
    _mm_maskmoveu_si128 ( _mm_set_epi8 (
            0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 ,
            sumY.m128i_u16 [ 7 ] / sumC.m128i_i16 [ 7 ] ,
            sumY.m128i_u16 [ 6 ] / sumC.m128i_i16 [ 6 ] ,
            sumY.m128i_u16 [ 5 ] / sumC.m128i_i16 [ 5 ] ,
            sumY.m128i_u16 [ 4 ] / sumC.m128i_i16 [ 4 ] ,
            sumY.m128i_u16 [ 3 ] / sumC.m128i_i16 [ 3 ] ,
            sumY.m128i_u16 [ 2 ] / sumC.m128i_i16 [ 2 ] ,
            sumY.m128i_u16 [ 1 ] / sumC.m128i_i16 [ 1 ] ,
            sumY.m128i_u16 [ 0 ] / sumC.m128i_i16 [ 0 ] ) ,
            _mm_set_epi32 ( 0 , 0 , 0x80808080 , 0x80808080 ) , (char*) d ) ;
}

static void SetFrame_YV12_SSE4 ( __m128i& sumY , __m128i& sumC , __m128i * d )
{
    _mm_storeu_si128 ( d , _mm_blendv_epi8 ( _mm_lddqu_si128 ( d ) ,
        _mm_set_epi8 (
            0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 ,
            _mm_extract_epi16 ( sumY , 7 ) / _mm_extract_epi16 ( sumC , 7 ) ,
            _mm_extract_epi16 ( sumY , 6 ) / _mm_extract_epi16 ( sumC , 6 ) ,
            _mm_extract_epi16 ( sumY , 5 ) / _mm_extract_epi16 ( sumC , 5 ) ,
            _mm_extract_epi16 ( sumY , 4 ) / _mm_extract_epi16 ( sumC , 4 ) ,
            _mm_extract_epi16 ( sumY , 3 ) / _mm_extract_epi16 ( sumC , 3 ) ,
            _mm_extract_epi16 ( sumY , 2 ) / _mm_extract_epi16 ( sumC , 2 ) ,
            _mm_extract_epi16 ( sumY , 1 ) / _mm_extract_epi16 ( sumC , 1 ) ,
            _mm_extract_epi16 ( sumY , 0 ) / _mm_extract_epi16 ( sumC , 0 ) ) ,
        _mm_set_epi32 ( 0 , 0 , 0x80808080 , 0x80808080 ) ) );
}

static void SetFrame_YUY2_SSE2 ( __m128i& sumY , __m128i& sumC , __m128i * d )
{
    _mm_maskmoveu_si128 ( _mm_set_epi16 (
            sumY.m128i_u16 [ 7 ] / sumC.m128i_i16 [ 7 ] ,
            sumY.m128i_u16 [ 6 ] / sumC.m128i_i16 [ 6 ] ,
            sumY.m128i_u16 [ 5 ] / sumC.m128i_i16 [ 5 ] ,
            sumY.m128i_u16 [ 4 ] / sumC.m128i_i16 [ 4 ] ,
            sumY.m128i_u16 [ 3 ] / sumC.m128i_i16 [ 3 ] ,
            sumY.m128i_u16 [ 2 ] / sumC.m128i_i16 [ 2 ] ,
            sumY.m128i_u16 [ 1 ] / sumC.m128i_i16 [ 1 ] ,
            sumY.m128i_u16 [ 0 ] / sumC.m128i_i16 [ 0 ] ) ,
            _mm_set1_epi16 ( 0x0080 ) , (char*) d ) ;
}

static void SetFrame_YUY2_SSE4 ( __m128i& sumY , __m128i& sumC , __m128i * d )
{
    _mm_storeu_si128 ( d , _mm_blendv_epi8 ( _mm_lddqu_si128 ( d ) ,
        _mm_set_epi16 (
            _mm_extract_epi16 ( sumY , 7 ) / _mm_extract_epi16 ( sumC , 7 ) ,
            _mm_extract_epi16 ( sumY , 6 ) / _mm_extract_epi16 ( sumC , 6 ) ,
            _mm_extract_epi16 ( sumY , 5 ) / _mm_extract_epi16 ( sumC , 5 ) ,
            _mm_extract_epi16 ( sumY , 4 ) / _mm_extract_epi16 ( sumC , 4 ) ,
            _mm_extract_epi16 ( sumY , 3 ) / _mm_extract_epi16 ( sumC , 3 ) ,
            _mm_extract_epi16 ( sumY , 2 ) / _mm_extract_epi16 ( sumC , 2 ) ,
            _mm_extract_epi16 ( sumY , 1 ) / _mm_extract_epi16 ( sumC , 1 ) ,
            _mm_extract_epi16 ( sumY , 0 ) / _mm_extract_epi16 ( sumC , 0 ) ) ,
        _mm_set1_epi16 ( 0x0080 ) ) );
}

//==============================================================================
// SetFrame
//==============================================================================
static DWORD WINAPI ThreadFunc_SSE2 ( LPVOID lpParameter )
{
    THREAD_PARAM2& p = * (THREAD_PARAM2*) lpParameter ;
    const __m128i sSelfWeight = _mm_set1_epi16 ( p.SelfWeight ) ;
    __m128i sumY ;
    __m128i sumC ;
    __m128i Yv ;
    __m128i AddMask ;
    __m128i AngV ;
    BYTE * d ;
    const int nDstU = ( p.IsYV12 ) ? 8 : 16 ;
    for ( int j = p.thr_num ; j < p.Height ; j += p.Threads )
    {
        d = p.pSrcY + j * p.SrcYPitch ;
        for ( int i = 0 ; i < p.Width ; i += 8 )
        {
            const int offset = i + p.CacheLOffset + ( j + max ( p.wSpan , 1 ) ) * p.CacheLSpan ;
            const int offsetCache = offset - max ( p.wSpan , 1 ) * p.CacheLSpan - p.wSpan ;
            const __m128i sY   = _mm_load_si128 ( (__m128i*) ( p.Y_FrameCache [ 0 ] + offset ) ) ;
            const __m128i sHag = _mm_load_si128 ( (__m128i*) ( p.H_FrameCache [ 0 ] + offset ) ) ;
            const __m128i sVag = _mm_load_si128 ( (__m128i*) ( p.V_FrameCache [ 0 ] + offset ) ) ;
            sumY = _mm_mullo_epi16 ( sY , sSelfWeight ) ;
            sumC = sSelfWeight ;
            for ( int dt = -p.tSpan ; dt <= p.tSpan ; dt ++ )
            {
                const __int16 * pY0j = p.Y_FrameCache [ dt ] + offsetCache ;
                const __int16 * pH0j = p.H_FrameCache [ dt ] + offsetCache ;
                const __int16 * pV0j = p.V_FrameCache [ dt ] + offsetCache ;
                for ( int dj = -p.wSpan ; dj <= p.wSpan ; dj ++ , pY0j += p.CacheLSpan , pH0j += p.CacheLSpan , pV0j += p.CacheLSpan )
                {
                    const __int16 * pY0i = pY0j ;
                    const __int16 * pH0i = pH0j ;
                    const __int16 * pV0i = pV0j ;
                    for ( int di = -p.wSpan ; di <= p.wSpan ; di ++ , pY0i ++ , pH0i ++ , pV0i ++ )
                    {
                        if ( ( dt | dj | di ) == 0 ) continue ;
                        Yv = _mm_loadu_si128 ( (const __m128i*) pY0i ) ;
                        AngV = _mm_add_epi16 ( _mm_srli_epi16 ( _mm_max_epi16 (
                            _mm_abs_epi16 ( _mm_sub_epi16 ( _mm_loadu_si128 ( (const __m128i*) pH0i ) , sHag ) ) ,
                            _mm_abs_epi16 ( _mm_sub_epi16 ( _mm_loadu_si128 ( (const __m128i*) pV0i ) , sVag ) ) ) ,
                            p.AngleWeight ) , _mm_set1_epi16 ( 1 ) ) ;
                        if ( _mm_movemask_epi8 ( _mm_cmpgt_epi16 ( AngV , _mm_set1_epi16 ( 0x7F ) ) ) )
                            AngV = _mm_and_si128 ( _mm_or_si128 ( _mm_cmpgt_epi16 ( AngV , _mm_set1_epi16 ( 0x7F ) ) , AngV ) , _mm_set1_epi16 ( 0x7F ) ) ;
                        AddMask = _mm_mullo_epi16 ( AngV , _mm_abs_epi16 ( _mm_sub_epi16 ( Yv , sY ) ) ) ;
                        if ( _mm_movemask_epi8 ( _mm_cmpgt_epi16 ( AddMask , _mm_set1_epi16 ( 0xFF ) ) ) )
                            AddMask = _mm_and_si128 ( _mm_or_si128 ( _mm_cmpgt_epi16 ( AddMask , Low16Mask ) , AddMask ) , Low16Mask ) ;
                        __m128i Weight = _mm_set_epi16 (
                            p.pW [ AddMask.m128i_i16 [ 7 ] ] ,
                            p.pW [ AddMask.m128i_i16 [ 6 ] ] ,
                            p.pW [ AddMask.m128i_i16 [ 5 ] ] ,
                            p.pW [ AddMask.m128i_i16 [ 4 ] ] ,
                            p.pW [ AddMask.m128i_i16 [ 3 ] ] ,
                            p.pW [ AddMask.m128i_i16 [ 2 ] ] ,
                            p.pW [ AddMask.m128i_i16 [ 1 ] ] ,
                            p.pW [ AddMask.m128i_i16 [ 0 ] ] ) ;
                        sumY = _mm_adds_epu16 ( sumY , _mm_mullo_epi16 ( Weight , Yv ) ) ;
                        sumC = _mm_add_epi16 ( sumC , Weight ) ;
                    }
                }
            }
            p.SetFrame ( _mm_adds_epu16 ( _mm_srli_epi16 ( sumC , 1 ) , sumY ) , sumC , (__m128i *) d ) ;
            d += nDstU ;
        }
    }
    return 0 ;
}

static DWORD WINAPI ThreadFunc_SSE4 ( LPVOID lpParameter )
{
    THREAD_PARAM2& p = * (THREAD_PARAM2*) lpParameter ;
    const __m128i sSelfWeight = _mm_set1_epi16 ( p.SelfWeight ) ;
    __m128i sumY ;
    __m128i sumC ;
    __m128i Yv ;
    __m128i AddMask ;
    __m128i AngV ;
    BYTE * d ;
    const int nDstU = ( p.IsYV12 ) ? 8 : 16 ;
    const int twSpan = max ( p.wSpan , 1 ) ;
    for ( int j = p.thr_num ; j < p.Height ; j += p.Threads )
    {
        d = p.pSrcY + j * p.SrcYPitch ;
        for ( int i = 0 ; i < p.Width ; i += 8 )
        {
            const int offset = i + p.CacheLOffset + ( j + twSpan ) * p.CacheLSpan ;
            const int offsetCache = offset - twSpan * p.CacheLSpan - p.wSpan ;
            const __m128i sY   = _mm_load_si128 ( (__m128i*) ( p.Y_FrameCache [ 0 ] + offset ) ) ;
            const __m128i sHag = _mm_load_si128 ( (__m128i*) ( p.H_FrameCache [ 0 ] + offset ) ) ;
            const __m128i sVag = _mm_load_si128 ( (__m128i*) ( p.V_FrameCache [ 0 ] + offset ) ) ;
            sumY = _mm_mullo_epi16 ( sY , sSelfWeight ) ;
            sumC = sSelfWeight ;
            for ( int dt = -p.tSpan ; dt <= p.tSpan ; dt ++ )
            {
                const __int16 * pY0j = p.Y_FrameCache [ dt ] + offsetCache ;
                const __int16 * pH0j = p.H_FrameCache [ dt ] + offsetCache ;
                const __int16 * pV0j = p.V_FrameCache [ dt ] + offsetCache ;
                for ( int dj = -p.wSpan ; dj <= p.wSpan ; dj ++ , pY0j += p.CacheLSpan , pH0j += p.CacheLSpan , pV0j += p.CacheLSpan )
                {
                    const __int16 * pY0i = pY0j ;
                    const __int16 * pH0i = pH0j ;
                    const __int16 * pV0i = pV0j ;
                    for ( int di = -p.wSpan ; di <= p.wSpan ; di ++ , pY0i ++ , pH0i ++ , pV0i ++ )
                    {
                        if ( ( dt | di | dj ) == 0 ) continue ;
                        Yv = _mm_lddqu_si128 ( (const __m128i*) pY0i ) ;
                        AngV = _mm_add_epi16 ( _mm_srli_epi16 ( _mm_max_epi16 (
                            _mm_abs_epi16 ( _mm_sub_epi16 ( _mm_lddqu_si128 ( (const __m128i*) pH0i ) , sHag ) ) ,
                            _mm_abs_epi16 ( _mm_sub_epi16 ( _mm_lddqu_si128 ( (const __m128i*) pV0i ) , sVag ) ) ) ,
                            p.AngleWeight ) , _mm_set1_epi16 ( 1 ) ) ;
                        if ( ! _mm_testz_si128 ( AngV , _mm_set1_epi16 ( (__int16) 0x0FF80 ) ) )
                            AngV = _mm_and_si128 ( _mm_or_si128 ( _mm_cmpgt_epi16 ( AngV , _mm_set1_epi16 ( 0x7F ) ) , AngV ) , _mm_set1_epi16 ( 0x7F ) ) ;
                        AddMask = _mm_mullo_epi16 ( AngV , _mm_abs_epi16 ( _mm_sub_epi16 ( Yv , sY ) ) ) ;
                        if ( ! _mm_testz_si128 ( AddMask , _mm_set1_epi16 ( (__int16) 0x0FF00 ) ) )
                            AddMask = _mm_and_si128 ( _mm_or_si128 ( _mm_cmpgt_epi16 ( AddMask , Low16Mask ) , AddMask ) , Low16Mask ) ;
                        __m128i Weight = _mm_set_epi16 (
                            p.pW [ _mm_extract_epi16 ( AddMask , 7 ) ] ,
                            p.pW [ _mm_extract_epi16 ( AddMask , 6 ) ] ,
                            p.pW [ _mm_extract_epi16 ( AddMask , 5 ) ] ,
                            p.pW [ _mm_extract_epi16 ( AddMask , 4 ) ] ,
                            p.pW [ _mm_extract_epi16 ( AddMask , 3 ) ] ,
                            p.pW [ _mm_extract_epi16 ( AddMask , 2 ) ] ,
                            p.pW [ _mm_extract_epi16 ( AddMask , 1 ) ] ,
                            p.pW [ _mm_extract_epi16 ( AddMask , 0 ) ] ) ;
                        sumY = _mm_adds_epu16 ( sumY , _mm_mullo_epi16 ( Weight , Yv ) ) ;
                        sumC = _mm_add_epi16 ( sumC , Weight ) ;
                    }
                }
            }
            p.SetFrame ( _mm_adds_epu16 ( _mm_srli_epi16 ( sumC , 1 ) , sumY ) , sumC , (__m128i *) d ) ;
            d += nDstU ;
        }
    }
    return 0 ;
}

