// $Date:: 2009-10-18 08:03:44 +0900#$
// $Rev: 16 $

#include "PNLM.h"
#include <math.h>
#include <crtdbg.h>
#include <malloc.h>

//#define PRINTTIME
#ifdef PRINTTIME
#include <stdio.h>
#include <mmsystem.h>
#pragma comment ( lib , "winmm.lib" )
static __int32 tcount = 0 ;
#endif

/**************************************************************************************************/
/* PNLM                                                                                      */
/**************************************************************************************************/
//==============================================================================
// constructor
//==============================================================================
PNLM::PNLM(PClip _child, int strength, int threshold1, int threshold2, IScriptEnvironment* env) : GenericVideoFilter(_child)
{
    this->Height    = this->vi.height ;
    this->Width     = this->vi.width ;
    
    // mCYxB0`20 ͈̔͂ŏB
    this->Strength  = strength ;

    // GbWo臒lB0`255 ͈̔͂ŏB
    this->Threshold1 = threshold1 ;

    // ڂ臒lB0`255 ͈̔͂ŏB
    this->Threshold2 = threshold2 ;

    // Zx->mCYlϊs̐B
    double H = - pow ( 10, (double) strength / -22.0) ;
    this->pW = ( (_weight*) malloc ( 511 * sizeof (_weight) ) ) + 255 ;
    for ( int i = -255 ; i < 256 ; i ++ )
    {
        this->pW [ i ].int16 [ 0 ] = (__int16) ( 8 * exp ( H * i * i ) + 0.5 );
        this->pW [ i ].int16 [ 1 ] = (__int16) ( 8 * exp ( H * i * i ) * i + 0.5 );
    }

    this->FrameCache = ( (BYTE**) malloc ( ( 2 * this->Threshold2 + 1 ) * sizeof (BYTE*) ) ) + this->Threshold2 ;
    memset ( this->FrameCache - this->Threshold2 , 0 , ( 2 * this->Threshold2 + 1 ) * sizeof (BYTE*) ) ;

    this->PreFrameNum = -2 ;
}

//==============================================================================
// destructor
//==============================================================================
PNLM::~PNLM()
{
    for ( int i = -this->Threshold2 ; i <= this->Threshold2  ; i ++ )
    {
        if ( this->FrameCache [ i ] ) free ( this->FrameCache [ i ] ) ;
    }
    free ( this->FrameCache - this->Threshold2 ) ;
    free ( this->pW - 255 ) ;
    //free ( this->pWc - 255 ) ;
    //free ( this->pWd - 255 ) ;

#ifdef PRINTTIME
    char strtime [ 256 ] ;
    _snprintf ( strtime , sizeof (strtime) , "%f\r\n" , ( (float) tcount ) / 1000 ) ;
    OutputDebugString ( strtime ) ;
#endif // PRINTTIME
}

//==============================================================================
// GetFrame
//==============================================================================
PVideoFrame PNLM::GetFrame (int n, IScriptEnvironment* env)
{
    PVideoFrame src = this->child->GetFrame ( n , env ) ;
    env->MakeWritable ( & src ) ;

    BYTE* pSrcY = src->GetWritePtr () ;

#ifdef PRINTTIME
    tcount -= timeGetTime () ;
#endif // PRINTTIME

    if ( this->PreFrameNum == -2 )
    {
        // ߂ GetFrame Ă΂ꂽƂ
        this->nSrcYPitch = src->GetPitch() ;
        for ( int i = -this->Threshold2 ; i <= this->Threshold2  ; i ++ )
        {
            this->FrameCache [ i ] = (BYTE*) malloc ( this->nSrcYPitch * this->Height ) ;
        }
    }

    if ( this->PreFrameNum + 1 != n )
    {
        // 
        for ( int i = -this->Threshold2 ; i < this->Threshold2  ; i ++ )
        {
            if ( n + i < 0 )
            {
                memcpy ( this->FrameCache [ i ] , this->child->GetFrame ( 0 , env )->GetReadPtr () ,
                    this->nSrcYPitch * this->Height );
            }
            else if ( n + i >= this->vi.num_frames )
            {
                memcpy ( this->FrameCache [ i ] , this->child->GetFrame ( this->vi.num_frames - 1 , env )->GetReadPtr () ,
                    this->nSrcYPitch * this->Height );
            }
            else
            {
                memcpy ( this->FrameCache [ i ] , this->child->GetFrame ( n + i , env )->GetReadPtr () ,
                    this->nSrcYPitch * this->Height );
            }
        }
    }
    else
    {
        BYTE* pTmp = this->FrameCache [ -this->Threshold2 ] ;
        memcpy ( this->FrameCache - this->Threshold2 , this->FrameCache - this->Threshold2 + 1 ,
            2 * this->Threshold2 * sizeof (BYTE*) ) ;
        this->FrameCache [ this->Threshold2 ] = pTmp ;
    }
    this->PreFrameNum = n ;

    if ( n + this->Threshold2 >= this->vi.num_frames )
    {
        memcpy ( this->FrameCache [ this->Threshold2 ] , this->child->GetFrame ( this->vi.num_frames - 1 , env )->GetReadPtr () ,
            this->nSrcYPitch * this->Height );
    }
    else
    {
        memcpy ( this->FrameCache [ this->Threshold2 ] , this->child->GetFrame ( n + this->Threshold2 , env )->GetReadPtr () ,
            this->nSrcYPitch * this->Height );
    }

    for ( int j = this->Threshold1 ; j < this->Height - this->Threshold1 ; j ++ )
    {
        for ( int i = this->Threshold1 ; i < this->Width - this->Threshold1 ; i ++ )
        {
            _weight sumv ;
            sumv.int32 = 0 ;
            const _weight * pv0 = this->pW - this->FrameCache [ 0 ][ i + this->nSrcYPitch * j ] ;
            for ( int dt = -this->Threshold2 ; dt <= this->Threshold2 ; dt ++ )
            {
                const BYTE * pv0j = this->FrameCache [ dt ] + i + ( j - this->Threshold1 ) * this->nSrcYPitch ;
                for ( int dj = -this->Threshold1 ; dj <= this->Threshold1 ; dj ++ , pv0j += this->nSrcYPitch )
                {
                    const BYTE * pv0i = pv0j - this->Threshold1 ;
                    for ( int di = -this->Threshold1 ; di <= this->Threshold1 ; di ++ , pv0i ++ )
                    {
                        sumv.int32 += pv0 [ * pv0i ].int32 ;
                    }
                }
            }
            pSrcY [ i + j * this->nSrcYPitch ] += sumv.int16 [ 1 ] / sumv.int16 [ 0 ] ;
        }
    }

#ifdef PRINTTIME
    tcount += timeGetTime () ;
#endif // PRINTTIME
    return src ;
}
